
import os, sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
import pytest
import pandas as pd
import dreem.util as util
import os
from dreem import clustering
from dreem.test import files_generator
from dreem.test.files_generator import test_files_dir, input_dir, prediction_dir, output_dir
import json 

module = 'clustering'

sample_name = 'test_set_1'
mode = 'light' # only uses the first sample

reads_partition = [
    [5000], [10000], [20000],
    [25000, 25000], [20000,80000], [10000,90000],
    [10000,20000,70000], [33333,33333,33334], [20000,40000,40000] 
]

half_sequence_length = [50, 100, 150]

unpaired_bases = {k:[int(k*l) for l in  [0.2, 0.4, 0.6]] for k in half_sequence_length}

shared_bases = {}
for k, v in unpaired_bases.items():
    shared_bases[k] = []
    for l in v:
        shared_bases[k].append([int(u*l) for u in  [0.2, 0.4, 0.6]])
        
sample_profile = {}
for r in reads_partition:
    for ac in half_sequence_length:
        for uc in unpaired_bases[ac]:
            for sc in shared_bases[ac][unpaired_bases[ac].index(uc)]:
                sample_profile['r{}_sl{}_ub{}_sb{}'.format(r, ac, uc, sc)] = {
                    'n_reads': r,
                    'n_AC': ac,
                    'n_unpaired': uc,
                    'n_shared': sc,
                    'path_bv': os.path.join(test_files_dir, 'input', module, sample_name, 'r{}_sl{}_ub{}_sb{}.orc'.format(r, ac, uc, sc)),
                    'path_json': os.path.join(test_files_dir, 'output', module, sample_name, 'r{}_sl{}_ub{}_sb{}'.format(r, ac, uc, sc))
                }

if mode == 'light':
    sample_profile = {}
    sample_profile['custom'] = {
                    'n_reads': [10000,10000],
                    'n_AC': 75,
                    'n_unpaired': 30,
                    'n_shared': 0,
                    'path_bv': os.path.join(test_files_dir, 'input', module, sample_name, 'custom.orc'),
                    'path_json': os.path.join(test_files_dir, 'output', module, sample_name),
                    'mu_unpaired': 0.05,
                    'mu_paired': 0.0
                }
    
module_input = os.path.join(input_dir, module)
module_expected = os.path.join(prediction_dir, module)
module_output =  os.path.join(output_dir, module)

inputs = ['clustering']

@pytest.mark.skip(reason="Too bugged")
def test_make_files():
    os.makedirs(os.path.join(test_files_dir, 'input', module, sample_name), exist_ok=True)
    os.makedirs(os.path.join(test_files_dir, 'expected_output', module, sample_name), exist_ok=True)
    files_generator.generate_files(sample_profile, module, inputs, [], test_files_dir, sample_name)
    files_generator.assert_files_exist(sample_profile, module, inputs, input_dir, sample_name)
    
@pytest.mark.skip(reason="Too bugged")
def test_run():
    clustering.run(
        input_dir =module_input,
        out_dir = module_output,
        num_runs= 3,
        min_iter=10,
        max_clusters=2,
        n_cpus=10
        )

@pytest.mark.skip(reason="Too bugged")
def test_assert_files_exist():
    assert os.path.exists(os.path.join(module_output, 'best_cluster_reads.json'))

@pytest.mark.skip(reason="not implemented")
def test_assess_performance():
    assert 1 == 0, 'not implemented'

if __name__ == '__main__':
    test_make_files()
    test_run()
    test_assert_files_exist()


import os
import shutil

from dreem.util.path import TEMP_DIR


# General parameters
DEFAULT_INTERLEAVED = False
SAM_HEADER = b"@"
SAM_ALIGN_SCORE = b"AS:i:"
SAM_EXTRA_SCORE = b"XS:i:"
FASTQ_REC_LENGTH = 4
DEFAULT_MIN_MAPQ = 30



class NgsFileBase(object):
    _operation_dir = ""

    def __init__(self,
                 root_dir: str,
                 ref_file: str,
                 sample: str) -> None:
        self._root_dir = root_dir
        self._ref_file = ref_file
        self._sample = sample
    
    @property
    def operation_dir(self):
        if not self._operation_dir:
            raise NotImplementedError
        return self._operation_dir
    
    @property
    def root_dir(self):
        return self._root_dir
    
    @property
    def ref_file(self):
        return self._ref_file
    
    @property
    def ref_prefix(self):
        return os.path.splitext(self.ref_file)[0]
    
    @property
    def ref_filename(self):
        return os.path.basename(self.ref_prefix)
    
    @property
    def sample(self):
        return self._sample
    
    def _get_dir(self, dest: str):
        return os.path.join(self.root_dir, dest,
                            self.operation_dir, self.sample)
    
    @property
    def output_dir(self):
        return self._get_dir(TEMP_DIR)
    
    def _make_output_dir(self):
        os.makedirs(self.output_dir, exist_ok=True)
    
    def run(self, *args, **kwargs):
        raise NotImplementedError
    
    def clean(self):
        shutil.rmtree(self.output_dir, ignore_errors=True)



import os
from multiprocessing import Pool

from dreem.util.cli import DEFAULT_INTERLEAVED_INPUT, DEFAULT_INTERLEAVE_OUTPUT, DEFAULT_TRIM, DEFAULT_NEXTSEQ_TRIM
from dreem.util.dflt import NUM_PROCESSES
from dreem.util.fa import FastaParser, FastaWriter
from dreem.util.fq import FastqAligner, FastqTrimmer, get_fastq_pairs, get_fastq_name
from dreem.util.star import starstarmap
from dreem.util.path import TEMP_DIR, try_remove
from dreem.util.seq import DNA
from dreem.util.xam import SamSorter, SamRemoveEqualMappers, SamOutputter, SamSplitter


def align_pipeline(out_dir: str, ref_file: str, sample: str, fastq: str,
             fastq2: str, trim: bool=DEFAULT_TRIM,
             interleaved_in: bool=DEFAULT_INTERLEAVED_INPUT,
             interleave_out: bool=DEFAULT_INTERLEAVE_OUTPUT,
             nextseq_trim: bool=DEFAULT_NEXTSEQ_TRIM):
    paired = interleaved_in or fastq2
    # Trim the FASTQ file.
    if trim:
        trimmer = FastqTrimmer(out_dir, ref_file, sample, paired, fastq, fastq2,
                               interleave_out=interleave_out)
        trimmer.run(nextseq_trim=nextseq_trim)
        fqs = trimmer.outputs
    else:
        fqs = [fastq]
        if fastq2:
            fqs.append(fastq2)

    # Align the FASTQ to the reference.
    aligner = FastqAligner(out_dir, ref_file, sample, paired, *fqs)
    aligner.run()
    trimmer.clean()
    # Remove equally mapping reads.
    rmequal = SamRemoveEqualMappers(out_dir, ref_file, sample, paired,
                                    aligner.sam_out)
    rmequal.run()
    aligner.clean()
    # Sort the SAM file and output a BAM file.
    sorter = SamSorter(out_dir, ref_file, sample, rmequal.sam_out)
    sorter.run()
    rmequal.clean()
    # Split the BAM file into one file for each reference.
    splitter = SamSplitter(out_dir, ref_file, sample, sorter.bam_out)
    splitter.run()
    sorter.clean()
    # Move the BAM files to the final output directory.
    bams = list()
    for bam in splitter.bams_out:
        outputter = SamOutputter(out_dir, ref_file, sample, bam)
        outputter.run()
        bams.append(outputter.bam_out)
    splitter.clean()
    return bams


def _dmplex_ref(out_dir: str, ref: bytes, seq: DNA, sample: str, fastq1: str,
                fastq2: str, **kwargs):
    """Run the alignment module.

    Aligns the reads to the reference genome and outputs one bam file perconstruct in the directory `output_path`, using `temp_path` as a temp directory.

    Parameters from args:
    -----------------------
    construct: str
        Name of the construct.
    sequence: str
        Sequence of the construct.    
    fastq1: str
        Path to the FASTQ file or list of paths to the FASTQ files, forward primer.
    fastq2: str
        Path to the FASTQ file or list of paths to the FASTQ files, reverse primer.
    output: str
        Path to the output folder (the sample).
    
    """

    # Write a temporary FASTA file for the reference.
    temp_dir = os.path.join(out_dir, TEMP_DIR, "alignment")
    temp_fasta_dir = os.path.join(temp_dir, "fasta")
    temp_fasta = os.path.join(temp_fasta_dir, f"{ref.decode()}.fasta")
    os.makedirs(temp_fasta_dir, exist_ok=True)
    
   # try:
    FastaWriter(temp_fasta, {ref: seq}).write()
    align_pipeline(out_dir, temp_fasta, sample, fastq1, fastq2, **kwargs)

   # finally:
   #     try_remove(temp_fasta)
   #     try:
   #         os.rmdir(temp_fasta_dir)
   #     except OSError:
   #         pass


def demultiplexed_fun(out_dir: str, fasta: str, sample: str, fastq: str,
                  fastq2: str, **kwargs):
    fq_dir = os.path.dirname(fastq)
    if fastq2:
        if fastq2.replace('_R2.','_R1.') != fastq:
            raise ValueError("fastq1 and fastq2 must be equal")
        #pairs = get_fastq_pairs(fq_dir)
        pairs = {get_fastq_name(fastq)[:-len('_R1')].encode('ascii'): (fastq, fastq2)}

    else:
        pairs = {get_fastq_name(fq): (os.path.join(fq_dir, fq), "")
                 for fq in os.listdir(fq_dir)}
    align_args = list()
    align_kwargs = list()
    for ref, seq in FastaParser(fasta).parse():
        try:
            fq1, fq2 = pairs[ref]
        except KeyError:
            pass
        else:
            arg = (out_dir, ref, seq, sample, fq1, fq2)
            align_args.append(arg)
            align_kwargs.append(kwargs)
            
    assert len(align_args) == len(align_kwargs)
    if align_args:
        n_procs = min(len(align_args), NUM_PROCESSES)
        with Pool(n_procs) as pool:
            starstarmap(pool.starmap, _dmplex_ref, align_args, align_kwargs)
            


from __future__ import annotations
import os
from pathlib import Path
import re
from typing import Any, Optional, Callable


# Define names of module directories.
MOD_DMX = "demultiplexing"
MOD_ALN = "alignment"
MOD_VEC = "vectoring"
MOD_CLS = "clustering"
MOD_AGG = "aggregation"
MODULES = (MOD_DMX, MOD_ALN, MOD_VEC, MOD_CLS, MOD_AGG)


# Define default values for path components.
DEF_OUT_ROOT = os.getcwd()
DEF_MODULE = ""
DEF_SAMPLE = ""
DEF_REF_NAME = ""
DEF_REGION_FIRST = -1
DEF_REGION_LAST = -1
DEF_MV_BATCH = -1
DEF_MV_REPORT = False


OUTPUT_DIR = "output"
TEMP_DIR = "temp"


# Path functions

def sanitize(path: str):
    return os.path.realpath(os.path.normpath(os.path.abspath(path)))


def switch_directory(old_path: str, new_dir: str):
    return os.path.join(new_dir, os.path.basename(old_path))


def try_remove(file: str):
    try:
        os.remove(file)
    except OSError:
        pass
    

class PathPart(object):
    def __init__(self, segment: str, require: bool=False) -> None:
        self._segment = segment
        if require:
            self.assert_exists()
    
    @property
    def path(self) -> Path:
        raise NotImplementedError
    
    @property
    def name(self) -> str:
        raise NotImplementedError
    
    @property
    def is_dir(self) -> bool:
        return True

    @property
    def is_file(self) -> bool:
        return not self.is_dir
    
    @property
    def exists(self) -> bool:
        if self.is_dir:
            return self.path.is_dir()
        if self.is_file:
            return self.path.is_file()
        assert False, f"Path '{self}' is neither directory nor file."
    
    def assert_exists(self):
        if not self.exists:
            raise FileNotFoundError(self.path)
    
    def __str__(self) -> str:
        return str(self.path)


class TopDir(PathPart):
    def __init__(self, path: str, require: bool = False) -> None:
        """
        Initialize instance of PathTopLevel to represent the top-level
        directory of a DREEM path. Top-level means the directory at and above
        which the file structure is not relevant to the operation of DREEM.
        For example, if your project's files are in the directory
          /home/me/projects/my_fav_rna
        and you tell DREEM to output all results to the directory
          /home/me/projects/my_fav_rna/results
        then the top-level directory would be
          /home/me/projects/my_fav_rna/results
        which would contain any of the following sub-directories
          /home/me/projects/my_fav_rna/results/demultiplexing
          /home/me/projects/my_fav_rna/results/alignment
          /home/me/projects/my_fav_rna/results/vectoring
          /home/me/projects/my_fav_rna/results/clustering
          /home/me/projects/my_fav_rna/results/aggregation
        
        Arguments
        ---------
        path: str
            Path (absolute or relative) of the top-level directory into which
            all outputs from DREEM are written
        
        require: bool (default: False)
            Whether the path must exist at the time this function is called
        
        Returns
        -------
        None
        """
        super().__init__(sanitize(path), require)
    
    @property
    def path(self):
        return Path(self._segment)


class PathSubPart(PathPart):
    def __init__(self, parent: PathPart, name: str, require: bool=False):
        """
        Initialize instance of PathSubPart to represent directories and files
        within a top-level directory. This class represents directories, and
        the subclass PathFilePart represents files.

        Arguments
        ---------
        parent: PathLevel
            The immediate parent of the directory represented by this instance
        
        name: str
            Base name of the directory represented by this instance (i.e. the
            last component of the path, everything after the last separator).
            May not contain any path separators.
        
        require: bool (default: False)
            Whether the full path (including the final component, name) must
            exist at the time this instance is initialized
        
        Returns
        -------
        None

        Raises
        ------
        ValueError
            if name contains any path separators or parent does not represent
            the path of a directory (whether or not that directory exists)
        """
        if os.sep in name:
            raise ValueError(f"name ({name}) may not contain '{os.sep}'")
        if not parent.is_dir:
            raise ValueError(f"{parent} is not a directory: cannot be parent.")
        self._parent = parent
        super().__init__(name, require)
    
    @property
    def parent(self):
        return self._parent
    
    @property
    def name(self):
        return self._segment
    
    @property
    def path(self):
        return self.parent.path.joinpath(self.name)

    def __getattribute__(self, name: str) -> Any:
        """
        Get an attribute of a PathSubPart instance. If the instance does not
        have the attribute, an AttributeError will be raised and caught.
        Then, the instance will check if its parent has the attribute.
        This search will continue up the chain of parents until either the
        attribute is found or the search reaches a TopDir, which neither
        implements this custom __getattribute__ method nor has a parent
        directory and will thus raise an uncaught AttributeError.
        This strategy of checking the parent for the attribute if the instance
        does not define it is designed to enable child directories to access
        attributes of their parents: for example, so a directory representing
        a region in a reference can determine which reference and sample the
        region belongs to.

        Warning: Using this method improperly can cause infinite recursion.
        
        Arguments
        ---------
        name: str
            Name of the attribute
        
        Returns
        -------
        any
            Value of the attribute for this instance, or if not found, for the
            closest parent with the attribute.
        
        Raises
        ------
        AttributeError
            if neither this instance nor any of its ancestors have the attribute
        """
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            return self.parent.__getattribute__(name)


class DreemPath(object):
    parent_type: type = type(None)
    
    @classmethod
    def parse(cls, path: str, require: bool) -> DreemPath:
        raise NotImplementedError


class DreemTopDir(TopDir, DreemPath):
    def __init__(self, path: str, require: bool=False) -> None:
        super().__init__(path, require)

    @property
    def top(self):
        return self

    @classmethod
    def parse(cls, path: str, require: bool=True):
        return cls(path, require)


class DreemSubPath(PathSubPart, DreemPath):
    def __init__(self, parent: PathPart, name: str, require: bool=False):
        if not isinstance(parent, self.parent_type):
            raise ValueError(
                f"Parent of {type(self)} must be {self.parent_type}, "
                f"but got {type(parent)}.")
        super().__init__(parent, name, require)
    
    @classmethod
    def parse(cls, path: str, require: bool=True):
        """
        Parse a raw string and return a DreemPath.
        """
        head, tail = os.path.split(sanitize(path))
        parent = cls.parent_type.parse(head, require)
        return cls(parent, tail, require)


class DreemFile(DreemSubPath):
    @property
    def is_dir(self):
        return False


class DreemOutDir(DreemSubPath):
    parent_type = DreemTopDir

    def __init__(self, parent: parent_type, name: str, require: bool=False):
        if name not in (OUTPUT_DIR, TEMP_DIR):
            raise ValueError(f"Invalid output directory name: '{name}'")
        super().__init__(parent, name, require)
    
    @property
    def is_temp(self):
        return self.name == TEMP_DIR


class DreemModuleDir(DreemSubPath):
    parent_type = DreemOutDir

    def __init__(self, parent: parent_type, name: str, require: bool=False):
        if name not in MODULES:
            raise ValueError(f"Invalid module name: '{name}'")
        super().__init__(parent, name, require)
    
    @property
    def module(self):
        return self


class DreemSampleDir(DreemSubPath):
    parent_type = DreemModuleDir

    @property
    def sample(self):
        return self


class DreemRefDir(DreemSubPath):
    parent_type = DreemSampleDir

    @property
    def ref(self):
        return self


class DreemRegionDir(DreemSubPath):
    parent_type = DreemRefDir
    name_format = "{first}-{last}"
    name_pattern = re.compile("^([0-9]+)-([0-9]+)$")

    def __init__(self, parent: PathPart, name: str, require: bool=False):
        super().__init__(parent, name, require)
        self._first, self._last = self.extract_bounds(self.name_pattern, name)

    @staticmethod
    def extract_bounds(pattern: re.Pattern[str], name: str):
        if not (match := pattern.match(name)):
            raise ValueError(f"Invalid region name: '{name}'")
        first, last = map(int, match.groups())
        if not 1 <= first <= last:
            raise ValueError(f"Invalid region bounds: {first}-{last}")
        return first, last
    
    @classmethod
    def assemble(cls, parent: PathPart, first: int, last: int):
        name = cls.name_format.format(first=first, last=last)
        return cls(parent, name)

    @property
    def region(self):
        """ Enable any children of this directory to access its attributes """
        return self
    
    @property
    def first(self):
        """ The first position in the region (1-indexed, inclusive) """
        return self._first
    
    @property
    def last(self):
        """ The last position in the region (1-indexed, inclusive) """
        return self._last
    
    def get_report_file(self):
        return DreemMpReportFile.assemble(self.parent, self.first, self.last)


class DreemMpReportFile(DreemFile):
    parent_type = DreemRefDir
    name_format = "{first}-{last}_report.txt"
    name_pattern = re.compile("^([0-9]+)-([0-9]+)_report.txt$")

    def __init__(self, parent: PathPart, name: str, require: bool=False):
        super().__init__(parent, name, require)
        if self.is_temp:
            raise ValueError(f"Report cannot be in temporary directory.")
        if (mod := self.module.name) != MOD_VEC:
            raise ValueError(f"Module must be '{MOD_VEC}', but got '{mod}'.")
        self._first, self._last = DreemRegionDir.extract_bounds(
            self.name_pattern, name)

    @classmethod
    def assemble(cls, parent: PathPart, first: int, last: int):
        name = cls.name_format.format(first=first, last=last)
        return cls(parent, name)
    
    @property
    def first(self):
        """ The first position in the region (1-indexed, inclusive) """
        return self._first

    @property
    def last(self):
        """ The last position in the region (1-indexed, inclusive) """
        return self._last
    
    def get_region_dir(self):
        return DreemRegionDir.assemble(self.parent, self.first, self.last)



import shlex
import subprocess,os

from typing import List


# Commands for external applications

BOWTIE2_CMD = "bowtie2"
BOWTIE2_BUILD_CMD = "bowtie2-build"
CUTADAPT_CMD = "cutadapt"
FASTQC_CMD = "fastqc"
SAMTOOLS_CMD = "samtools"


# Command utility functions

def run_cmd(args: List[str], shell: bool = False):
    os.system(' '.join(args))#, check=True, shell=shell)
    cmd = shlex.join(args)
    return cmd



from typing import Dict, Set

from dreem.util.seq import DNA


class FastaIO(object):
    __slots__ = ["_path", "_refs"]

    defsymbol = b">"
    deftrunc = len(defsymbol)

    def __init__(self, path: str):
        self._path = path


class FastaParser(FastaIO):
    def __init__(self, path: str):
        super().__init__(path)
        self._refs: Set[bytes] = set()

    @classmethod
    def _parse_fasta_record(cls, fasta, line: bytes):
        if not line.startswith(cls.defsymbol):
            raise ValueError("FASTA definition line does not start with "
                             f"'{cls.defsymbol.decode()}'")
        name = line.split()[0][cls.deftrunc:]
        seq = bytearray()
        while (line := fasta.readline()) and not line.startswith(cls.defsymbol):
            seq.extend(line.rstrip())
        seq = DNA(bytes(seq))
        return line, name, seq

    def parse(self):
        with open(self._path, "rb") as f:
            line = f.readline()
            while line:
                line, name, seq = self._parse_fasta_record(f, line)
                if name in self._refs:
                    raise ValueError(
                        f"Duplicate entry in {self._path}: '{name.decode()}'")
                self._refs.add(name)
                yield name, seq


class FastaWriter(FastaIO):
    def __init__(self, path: str, refs: Dict[bytes, DNA]):
        super().__init__(path)
        self._refs = refs
    
    def write(self):
        with open(self._path, "wb") as f:
            for ref, seq in self._refs.items():
                f.write(b"".join((self.defsymbol, ref, b"\n", seq, b"\n")))
                

def parse_fasta(path: str):
    return FastaParser(path).parse()



from collections import defaultdict
from multiprocessing import Pool
import os

from typing import List, Optional, Tuple

from dreem.util.cmd import FASTQC_CMD, CUTADAPT_CMD, BOWTIE2_CMD, \
    BOWTIE2_BUILD_CMD, run_cmd
from dreem.util.dflt import NUM_PROCESSES, PHRED_ENCODING
from dreem.util.ngs import NgsFileBase
from dreem.util.seq import get_diffs
from dreem.util.cli import DEFAULT_MIN_BASE_QUALITY, DEFAULT_ILLUMINA_ADAPTER, DEFAULT_MIN_OVERLAP, DEFAULT_MAX_ERROR, DEFAULT_INDELS, DEFAULT_NEXTSEQ_TRIM, DEFAULT_DISCARD_TRIMMED, DEFAULT_DISCARD_UNTRIMMED, DEFAULT_MIN_LENGTH, DEFAULT_SCORE_MIN
from dreem.util.cli import DEFAULT_LOCAL, DEFAULT_UNALIGNED, DEFAULT_DISCORDANT, DEFAULT_MIXED, DEFAULT_DOVETAIL, DEFAULT_CONTAIN, DEFAULT_FRAG_LEN_MIN, DEFAULT_FRAG_LEN_MAX, DEFAULT_N_CEILING, DEFAULT_SEED_INTERVAL, DEFAULT_GAP_BAR, DEFAULT_SEED_SIZE, DEFAULT_EXTENSIONS, DEFAULT_RESEED, DEFAULT_PADDING, DEFAULT_ALIGN_THREADS, MATCH_BONUS, MISMATCH_PENALTY, N_PENALTY, REF_GAP_PENALTY, READ_GAP_PENALTY, IGNORE_QUALS
from dreem.util.cli import DEFAULT_INTERLEAVE_OUTPUT, FASTQ2


# FastQC parameters
DEFAULT_EXTRACT = False


def get_fastq_name(fastq: str, fastq2: str=""):
    exts = (".fq", ".fastq")
    reads = ("mate", "r")
    base = os.path.basename(fastq)
    counts = {ext: count for ext in exts if (count := base.count(ext))}
    if not counts:
        raise ValueError(f"File '{fastq}' had no FASTQ extension.")
    if sum(counts.values()) > 1:
        raise ValueError(f"File '{fastq}' had multiple FASTQ extensions.")
    name = base[:base.index(list(counts)[0])]
    if fastq2:
        name2 = get_fastq_name(fastq2)
        diffs = get_diffs(name, name2)
        if len(diffs) != 1:
            raise ValueError("FASTQ names must differ by exactly 1 character.")
        idx = diffs[0]
        if name[idx] != "1" or name2[idx] != "2":
            raise ValueError("FASTQs must be named '1' and '2', respectively.")
        name = name[:idx]
        for read in reads:
            if name.rstrip("_").lower().endswith(read):
                name = name[:name.lower().rindex(read)]
                break
        name = name.rstrip("_")
    return name


def get_fastq_dir(fastq: str, fastq2: Optional[str]=None):
    fq_dir = os.path.dirname(fastq)
    if fastq2 and os.path.dirname(fastq2) != fq_dir:
        raise ValueError("FASTQs are not in the same directory.")
    return os.path.basename(fq_dir)


def get_fastq_pairs(fq_dir: str):
    lengths = defaultdict(set)
    for file in os.listdir(fq_dir):
        lengths[len(file)].add(file)
    pairs = dict()
    for fq_files in lengths.values():
        while fq_files:
            file1 = fq_files.pop()
            diffs = defaultdict(set)
            for file2 in fq_files:
                diffs[len(get_diffs(file1, file2))].add(file2)
            try:
                file2 = diffs[1].pop()
            except KeyError:
                raise ValueError(f"Found no mate for {file1} in {fq_dir}")
            if diffs[1]:
                raise ValueError(f"Found ≥2 mates for {file1} in {fq_dir}")
            fq_files.pop(file2)
            pair = (os.path.join(fq_dir, file1), os.path.join(fq_dir, file2))
            pairs[get_fastq_name(file1, file2)] = pair
    return pairs


class FastqBase(NgsFileBase):
    def __init__(self,
                 root_dir: str,
                 ref_file: str,
                 sample: str,
                 paired: bool,
                 fastq: str,
                 fastq2: str=FASTQ2,
                 interleave_out: bool=DEFAULT_INTERLEAVE_OUTPUT,
                 encoding: int = PHRED_ENCODING) -> None:
        if fastq2 and not paired:
            raise ValueError("fastq2 can only be given if reads are paired")
        super().__init__(root_dir, ref_file, sample)
        self._fq_in = fastq
        self._fq2_in = fastq2
        self._paired = paired
        self._interleave_out = interleave_out
        self._encoding = encoding
        self._name = get_fastq_name(fastq, fastq2)
    
    @property
    def fq_in(self):
        return self._fq_in
    
    @property
    def fq1_in(self):
        return self._fq_in
        
    @property
    def fq2_in(self):
        return self._fq2_in

    @property
    def paired(self):
        return self._paired

    @property
    def interleaved_in(self):
        return self.paired and not self.fq2_in
    
    @property
    def interleave_out(self):
        return self._interleave_out and self.paired

    @property
    def encoding(self):
        return self._encoding
    
    @property
    def encoding_arg(self):
        return f"--phred{self.encoding}"
    
    @property
    def name(self):
        return self._name
    
    @property
    def inputs(self):
        fqs = [self.fq_in]
        if self.fq2_in:
            fqs.append(self.fq2_in)
        return fqs
    
    @property
    def outputs(self):
        if self.paired and not self.interleave_out:
            fqs = [f"{self.name}_R{r}" for r in ("1", "2")]
        else:
            fqs = [self.name]
        return [os.path.join(self.output_dir, f"{fq}.fq") for fq in fqs]
    
    @staticmethod
    def _qc(files: List[str], extract: bool):
        cmd = [FASTQC_CMD]
        if extract:
            cmd.append("--extract")
        cmd.extend(files)
        run_cmd(cmd)
    
    def qc_inputs(self, extract: bool=DEFAULT_EXTRACT):
        return self._qc(self.inputs, extract)

    def qc_outputs(self, extract: bool=DEFAULT_EXTRACT):
        return self._qc(self.outputs, extract)


class FastqTrimmer(FastqBase):
    _operation_dir = "alignment/1_trim"

    def _cutadapt(self,
                  qual1: int=DEFAULT_MIN_BASE_QUALITY,
                  qual2: int=0,
                  adapters15: Tuple[str]=(),
                  adapters13: Tuple[str]=(DEFAULT_ILLUMINA_ADAPTER,),
                  adapters25: Tuple[str]=(),
                  adapters23: Tuple[str]=(DEFAULT_ILLUMINA_ADAPTER,),
                  min_overlap: int=DEFAULT_MIN_OVERLAP,
                  max_error: float=DEFAULT_MAX_ERROR,
                  indels: bool=DEFAULT_INDELS,
                  nextseq_trim: bool=DEFAULT_NEXTSEQ_TRIM,
                  discard_trimmed: bool=DEFAULT_DISCARD_TRIMMED,
                  discard_untrimmed: bool=DEFAULT_DISCARD_UNTRIMMED,
                  min_length: bool=DEFAULT_MIN_LENGTH,
                  cores: int=NUM_PROCESSES):
        cmd = [CUTADAPT_CMD]
        if cores >= 0:
            cmd.extend(["--cores", str(cores)])
        if nextseq_trim:
            if qual1 > 0:
                cmd.extend(["--nextseq-trim", str(qual1)])
        else:
            if qual1 > 0:
                cmd.extend(["-q", str(qual1)])
            if qual2 > 0:
                cmd.extend(["-Q", str(qual2)])
        adapters = {"g": adapters15, "a": adapters13,
                    "G": adapters25, "A": adapters23}
        for arg, adapter in adapters.items():
            if adapter and (self.paired or arg.islower()):
                for adapt in adapter:
                    cmd.extend([f"-{arg}", adapt])
        if min_overlap >= 0:
            cmd.extend(["-O", str(min_overlap)])
        if max_error >= 0:
            cmd.extend(["-e", str(max_error)])
        if not indels:
            cmd.append("--no-indels")
        if discard_trimmed:
            cmd.append("--discard-trimmed")
        if discard_untrimmed:
            cmd.append("--discard-untrimmed")
        if min_length:
            cmd.extend(["-m", str(min_length)])
        cmd.extend(["--report", "minimal"])
        if self.interleaved_in or self.interleave_out:
            cmd.append("--interleaved")
        for flag, output in zip(["-o", "-p"], self.outputs):
            cmd.extend([flag, output])
        cmd.extend(self.inputs)
        run_cmd(cmd)
    
    def run(self, **kwargs):
        print(f"\nTrimming Adapters from {self.fq_in}\n")
        self._make_output_dir()
        self._cutadapt(**kwargs)


class FastqAligner(FastqBase):
    _operation_dir = "alignment/2_align"

    @property
    def sam_out(self):
        return os.path.join(self.output_dir, f"{self.ref_filename}.sam")
    
    @property
    def outputs(self):
        return [self.sam_out]

    def _bowtie2_build(self):
        """
        Build an index of a reference genome using Bowtie 2.
        :param ref: (str) path to the reference genome FASTA file
        :return: None
        """
        cmd = [BOWTIE2_BUILD_CMD, "-q", self.ref_file, self.ref_prefix]
        run_cmd(cmd)
    
    def _bowtie2(self,
                 local=DEFAULT_LOCAL,
                 unaligned=DEFAULT_UNALIGNED,
                 discordant=DEFAULT_DISCORDANT,
                 mixed=DEFAULT_MIXED,
                 dovetail=DEFAULT_DOVETAIL,
                 contain=DEFAULT_CONTAIN,
                 score_min=DEFAULT_SCORE_MIN,
                 frag_len_min=DEFAULT_FRAG_LEN_MIN,
                 frag_len_max=DEFAULT_FRAG_LEN_MAX,
                 n_ceil=DEFAULT_N_CEILING,
                 gap_bar=DEFAULT_GAP_BAR,
                 seed_size=DEFAULT_SEED_SIZE,
                 seed_interval=DEFAULT_SEED_INTERVAL,
                 extensions=DEFAULT_EXTENSIONS,
                 reseed=DEFAULT_RESEED,
                 padding=DEFAULT_PADDING,
                 threads=DEFAULT_ALIGN_THREADS):
        cmd = [BOWTIE2_CMD]
        if self.interleaved_in:
            cmd.extend(["--interleaved", self.fq_in])
        elif self.paired:
            cmd.extend(["-1", self.fq1_in, "-2", self.fq2_in])
        else:
            cmd.extend(["-U", self.fq_in])
        cmd.extend(["-x", self.ref_prefix])
        cmd.extend(["-S", self.sam_out])
        cmd.append(self.encoding_arg)
        cmd.append("--xeq")
        cmd.extend(["--ma", MATCH_BONUS])
        cmd.extend(["--mp", MISMATCH_PENALTY])
        cmd.extend(["--np", N_PENALTY])
        cmd.extend(["--rfg", REF_GAP_PENALTY])
        cmd.extend(["--rdg", READ_GAP_PENALTY])
        if score_min:
            cmd.extend(["--score-min", score_min])
        if local:
            cmd.append("--local")
        if not unaligned:
            cmd.append("--no-unal")
        if not discordant:
            cmd.append("--no-discordant")
        if not mixed:
            cmd.append("--no-mixed")
        if dovetail:
            cmd.append("--dovetail")
        if not contain:
            cmd.append("--no-contain")
        if frag_len_min:
            cmd.extend(["-I", str(frag_len_min)])
        if frag_len_max:
            cmd.extend(["-X", str(frag_len_max)])
        if n_ceil:
            cmd.extend(["--n-ceil", n_ceil])
        if gap_bar:
            cmd.extend(["--gbar", str(gap_bar)])
        if seed_size:
            cmd.extend(["-L", str(seed_size)])
        if seed_interval:
            cmd.extend(["-i", str(seed_interval)])
        if extensions:
            cmd.extend(["-D", str(extensions)])
        if reseed:
            cmd.extend(["-R", str(reseed)])
        if padding:
            cmd.extend(["--dpad", str(padding)])
        if threads:
            cmd.extend(["-p", str(threads)])
        if IGNORE_QUALS:
            cmd.append("--ignore-quals")
        run_cmd(cmd)
    
    def run(self, **kwargs):
        print(f"\nAligning Reads {self.fq_in} to Reference {self.ref_file}\n")
        self._make_output_dir()
        self._bowtie2_build()
        self._bowtie2(**kwargs)


'''
primer1 = "CAGCACTCAGAGCTAATACGACTCACTATA"
primer1rc = "TATAGTGAGTCGTATTAGCTCTGAGTGCTG"
primer2 = "TGAAGAGCTGGAACGCTTCACTGA"
primer2rc = "TCAGTGAAGCGTTCCAGCTCTTCA"
adapters5 = (primer1, primer2rc)
adapters3 = (primer2, primer1rc)
'''



from __future__ import annotations
from functools import cached_property, wraps
from io import BufferedReader
import os
from typing import Optional

from dreem.util.util import name_temp_file, SAMTOOLS_CMD, run_cmd, try_remove
from dreem.vector.vector import *


def _requires_open(func: function):
    @wraps(func)
    def wrapper(self: SamViewer, *args, **kwargs):
        if self._sam_file is None:
            raise ValueError(f"Function '{func.__name__}' requires "
                             "the SAM file to have been opened.")
        return func(self, *args, **kwargs)
    return wrapper


def _reset_seek(func: function):
    @wraps(func)
    @_requires_open
    def wrapper(self: SamViewer, *args, **kwargs):
        prev_pos = self._sam_file.tell()
        ret = func(self, *args, **kwargs)
        self._sam_file.seek(prev_pos)
        return ret
    return wrapper


def _range_of_records(func: function):
    @wraps(func)
    @_reset_seek
    def wrapper(self: SamViewer, start: int, stop: int):
        self._sam_file.seek(start)
        records = iter(func(self))
        if stop is None:
            return records
        else:
            while True:
                if self._sam_file.tell() < stop:
                    yield next(records)
                else:
                    assert self._sam_file.tell() == stop
                    break
    return wrapper


class SamViewer(object):
    def __init__(self, xam_path: str, ref_name: bytes, first: int, last: int,
                 spanning: bool, make: bool = True, remove: bool = True):
        self._xam_in = xam_path
        self._ref_name = ref_name
        self._first = first
        self._last = last
        self._spanning = spanning
        self._make = make
        self._remove = remove
        self._sam_temp: str = ""
        self._sam_file: Optional[BufferedReader] = None
        self._paired: Optional[bool] = None
    
    @property
    def input_dir(self):
        return os.path.dirname(self._xam_in)
    
    @property
    def input_name(self):
        return os.path.splitext(os.path.basename(self._xam_in))[0]
    
    @property
    def ref_coords(self):
        return f"{self._ref_name.decode()}:{self._first}-{self._last}"
    
    @property
    def sam_subset(self):
        if not self._sam_temp:
            raise ValueError("Not currently viewing SAM subset file")
        return self._sam_temp
    
    def __enter__(self):
        # Convert the BAM file to a temporary SAM file
        
        # FIXME !!!
        # Use functions defined in fastq.py to do these conversions elegantly
        # Write the SAM files to the temporary directory instead of the same
        # directory with the BAM file (in case that directory is read-only)

        if self._make:
            cmd = [SAMTOOLS_CMD, "index", self._xam_in]
            run_cmd(cmd)

            sam_view = name_temp_file(dirname=self.input_dir,
                                    prefix=f"{self.input_name}_nosort_",
                                    suffix=".sam")
            cmd = [SAMTOOLS_CMD, "view", "-h", "-o", sam_view,
                self._xam_in, self.ref_coords]
            run_cmd(cmd)

            self._sam_temp = name_temp_file(dirname=self.input_dir,
                                            prefix=f"{self.input_name}_",
                                            suffix=".sam")
            cmd = [SAMTOOLS_CMD, "sort", "-n", "-o", self._sam_temp, sam_view]
            run_cmd(cmd)
            try_remove(sam_view)
            self._sam_file = open(self._sam_temp, "rb")
        else:
            self._sam_file = open(self._xam_in, "rb")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._sam_file.close()
        self._sam_file = None
        if self._remove:
            try_remove(self._sam_temp)
        self._sam_temp = ""

    @_requires_open
    def _seek_beginning(self):
        self._sam_file.seek(0)
    
    @cached_property
    @_reset_seek
    def _rec1_pos(self):
        self._seek_beginning()
        while (line := self._sam_file.readline()).startswith(SAM_HEADER):
            pass
        return self._sam_file.tell() - len(line)
    
    @_requires_open
    def _seek_rec1(self):
        self._sam_file.seek(self._rec1_pos)
        
    @cached_property
    @_reset_seek
    def paired(self):
        if self._paired is None:
            self._seek_rec1()
            first_line = self._sam_file.readline()
            self._paired = (SamRead(first_line).flag.paired if first_line
                            else False)
        return self._paired
    
    @_range_of_records
    def _get_records_single(self):
        while read := SamRead(self._sam_file.readline()):
            assert not read.flag.paired
            yield SamRecord(read)

    @_range_of_records
    def _get_records_paired_flexible(self):
        prev_read: Optional[SamRead] = None
        while line := self._sam_file.readline():
            read = SamRead(line)
            assert read.flag.paired
            if prev_read:
                # The previous read has not yet been yielded
                if prev_read.qname == read.qname:
                    # The current read is the mate of the previous read
                    if prev_read.flag.first:
                        yield SamRecord(prev_read, read)
                    else:
                        yield SamRecord(read, prev_read)
                    prev_read = None
                else:
                    # The previous read is paired, but its mate is not in
                    # the SAM file. This situation can happen when only the
                    # reads mapping to a specific region are exported to a
                    # SAM file and only one mate overlaps that region.
                    yield SamRecord(prev_read)
                    # Save the current read so that if its mate is the next
                    # read, it will be returned as a pair.
                    prev_read = read
            else:
                # Save the current read so that if its mate is the next
                # read, it will be returned as a pair.
                prev_read = read
        if prev_read:
            # In case the last record
            yield SamRecord(prev_read)
    
    @_range_of_records
    def _get_records_paired_strict(self):
        while line := self._sam_file.readline():
            yield SamRecord(SamRead(line), SamRead(self._sam_file.readline()))
    
    def get_records(self, start: int, stop: int):
        if self.paired:
            if self._spanning:
                records = self._get_records_paired_strict
            else:
                records = self._get_records_paired_flexible
        else:
            records = self._get_records_single
        return records(start, stop)
    
    @_reset_seek
    def get_batch_indexes(self, batch_size: int):
        if batch_size <= 0:
            raise ValueError("batch_size must be a positive integer")
        n_skip = (self.paired + 1) * (batch_size - 1)
        self._seek_rec1()
        while True:
            yield self._sam_file.tell()
            try:
                line = next(self._sam_file)
            except StopIteration:
                break
            else:
                for _, line in zip(range(n_skip), self._sam_file):
                    pass
                if self.paired:
                    # Compare the current and next query names.
                    try:
                        qname = line.split()[0]
                    except IndexError:
                        pass
                    else:
                        line_next = self._sam_file.readline()
                        try:
                            qname_next = line_next.split()[0]
                        except IndexError:
                            pass
                        else:
                            if qname != qname_next:
                                # If the current and next query names differ
                                # (the lines do not come from two paired mates),
                                # then backtrack to the beginning of line_next.
                                self._sam_file.seek(-len(line_next), 1)



import itertools
from io import BufferedReader
import os
import re

from dreem.util.cmd import SAMTOOLS_CMD, run_cmd
from dreem.util.dflt import BUFFER_LENGTH
from dreem.util.fa import FastaParser
from dreem.util.ngs import NgsFileBase, SAM_ALIGN_SCORE, SAM_EXTRA_SCORE, SAM_HEADER
from dreem.util.path import OUTPUT_DIR, switch_directory


SAM_EXT = ".sam"
BAM_EXT = ".bam"


class SamBase(NgsFileBase):
    def __init__(self,
                 root_dir: str,
                 ref_file: str,
                 sample: str,
                 xam_file: str) -> None:
        super().__init__(root_dir, ref_file, sample)
        self._ref_names = [ref for ref, _ in FastaParser(ref_file).parse()]
        self._xam_in = xam_file
    
    @property
    def refs(self):
        return self._ref_names
    
    @property
    def xam_in(self):
        return self._xam_in
    
    @property
    def name(self):
        return os.path.splitext(os.path.basename(self.xam_in))[0]
    
    @property
    def sam_name(self):
        return f"{self.name}.sam"

    @property
    def bam_name(self):
        return f"{self.name}{BAM_EXT}"
    
    @property
    def sam_out(self):
        return os.path.join(self.output_dir, self.sam_name)
    
    @property
    def bam_out(self):
        return os.path.join(self.output_dir, self.bam_name)
    
    def _get_bam_split(self, ref: bytes):
        return os.path.join(self.output_dir, f"{ref}{BAM_EXT}")
    
    @staticmethod
    def _index_bam(bam_file: str):
        cmd = [SAMTOOLS_CMD, "index", bam_file]
        run_cmd(cmd)

    def index_bam_in(self):
        self._index_bam(self.xam_in)
    
    def index_bam_out(self):
        self._index_bam(self.bam_out)


class SamRemoveEqualMappers(SamBase):
    _operation_dir = "alignment/3_rem"

    pattern_a = re.compile(SAM_ALIGN_SCORE + rb"(\d+)")
    pattern_x = re.compile(SAM_EXTRA_SCORE + rb"(\d+)")

    def __init__(self, root_dir: str, ref_file: str, sample: str,
                 paired: bool, xam_file: str) -> None:
        super().__init__(root_dir, ref_file, sample, xam_file)
        self._paired = paired
    
    @property
    def paired(self):
        return self._paired

    @staticmethod
    def get_score(line: bytes, ptn: re.Pattern[bytes]):
        return (float(match.groups()[0])
                if (match := ptn.search(line)) else None)
    
    @classmethod
    def is_best_alignment(cls, line: bytes):
        return ((score_x := cls.get_score(line, cls.pattern_x)) is None
                or score_x < cls.get_score(line, cls.pattern_a))
    
    def _iter_paired(self, sam: BufferedReader, line: bytes):
        for line2 in sam:
            if self.is_best_alignment(line) or self.is_best_alignment(line2):
                yield b"".join((line, line2))
            line = sam.readline()
    
    def _iter_single(self, sam: BufferedReader, line: bytes):
        while line:
            if self.is_best_alignment(line):
                yield(line)
            line = sam.readline()

    def _remove_equal_mappers(self, buffer_length=BUFFER_LENGTH):
        with open(self.xam_in, "rb") as sami, open(self.sam_out, "wb") as samo:
            # Copy the header from the input to the output SAM file.
            while (line := sami.readline()).startswith(SAM_HEADER):
                samo.write(line)
            iter_sam = self._iter_paired if self.paired else self._iter_single
            lines = iter_sam(sami, line)
            while text := b"".join(itertools.islice(lines, buffer_length)):
                samo.write(text)
    
    def run(self):
        print("\nRemoving Reads Mapping Equally to Multiple Locations in "
              f"{self.xam_in}\n")
        self._make_output_dir()
        self._remove_equal_mappers()


class SamSorter(SamBase):
    _operation_dir = "alignment/4_sort"

    def _sort(self, name: bool = False):
        cmd = [SAMTOOLS_CMD, "sort"]
        if name:
            cmd.append("-n")
        cmd.extend(["-o", self.bam_out, self.xam_in])
        run_cmd(cmd)
    
    def run(self, name: bool = False):
        print(f"\nSorting {self.xam_in} by Reference and Coordinate\n")
        self._make_output_dir()
        self._sort()


class SamSplitter(SamBase):
    _operation_dir = "alignment/5_split"

    def _get_bam_out_ref(self, ref: bytes):
        return os.path.join(self.output_dir, f"{ref.decode()}{BAM_EXT}")
    
    @property
    def bams_out(self):
        return list(map(self._get_bam_out_ref, self.refs))
    
    @property
    def bam_out(self):
        raise NotImplementedError
    
    def _output_bam_ref(self, ref: bytes, bam_out: str):
        cmd = [SAMTOOLS_CMD, "view", "-b", "-o", bam_out,
               self.xam_in, ref.decode()]
        run_cmd(cmd)
    
    def _split_bam(self):
        self.index_bam_in()
        for ref, bam in zip(self.refs, self.bams_out):
            self._output_bam_ref(ref, bam)
    
    def run(self):
        print(f"\nSplitting {self.xam_in} into Individual References\n")
        self._make_output_dir()
        self._split_bam()


class SamOutputter(SamBase):
    _operation_dir = "alignment"

    @property
    def output_dir(self):
        return self._get_dir(OUTPUT_DIR)
    
    @property
    def bam_out(self):
        return switch_directory(self.xam_in, self.output_dir)
    
    def run(self):
        print(f"\nOutputting Cleaned BAM files to {self.output_dir}\n")
        self._make_output_dir()
        os.rename(self.xam_in, self.bam_out)
